import subprocess


# subprocess

def call_subprocess(command: str) -> bool:
    try:
        return (
            bool(
                subprocess.call(
                    command,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.STDOUT,
                    shell=True,
                )
            )
            == 0
        )
    except subprocess.CalledProcessError:
        return False

def check_subprocess_output(command: str) -> str:
    try:
        return subprocess.check_output(
            command, stderr=subprocess.STDOUT, shell=True, text=True
        )
    except subprocess.CalledProcessError:
        return ""

# Boot config (/boot/config.txt) - deprecated and not used anymore but can be useful

def read_config(config_file: str) -> dict[str, str]:
    config = {}

    with open(config_file, "r") as f:
        for line in f.readlines():
            line = line.strip()
            if line and not line.startswith("#"):
                splits = line.split("=", 1)  # Split only on the first '='
                if len(splits) == 2:
                    config[splits[0].strip()] = splits[1].strip()

    return config

def write_config(config: dict, config_file: str):
    with open(config_file, "w") as f:
        for k, v in config.items():
            f.write(f"{k}={v}\n")

def set_config_value(key: str, value: str, config_file: str):
    config = read_config(config_file)
    config[key] = value
    write_config(config, config_file)



interfaces_path = r"/etc/network/interfaces"
default_hostapd_path = r"/etc/default/hostapd"
hostapd_conf_path = r"/etc/hostapd/hostapd.conf"
dnsmaq_conf_path = r"/etc/dnsmasq.conf"
wpa_supplicant_path = r"/etc/wpa_supplicant/wpa_supplicant.conf"
hosts_path = r"/etc/hosts"

network_interfaces_content = """source-directory /etc/network/interfaces.d

auto lo
auto eth0
auto wlan0
auto ap0

iface eth0 inet dhcp
iface lo inet loopback

allow-hotplug ap0
iface ap0 inet static
    address 192.168.10.1
    netmask 255.255.255.0
    hostapd /etc/hostapd/hostapd.conf

allow-hotplug wlan0
iface wlan0 inet manual
    wpa-roam /etc/wpa_supplicant/wpa_supplicant.conf

"""

default_hostapd_content = """DAEMON_CONF="/etc/hostapd/hostapd.conf"
"""

hostapd_conf_content = """ctrl_interface=/var/run/hostapd
ctrl_interface_group=0
interface=ap0
driver=nl80211
ssid=MajorDom Hub
hw_mode=g
channel=11
wmm_enabled=0
macaddr_acl=0
auth_algs=1
wpa=2
wpa_passphrase=majordom
wpa_key_mgmt=WPA-PSK
wpa_pairwise=TKIP CCMP
rsn_pairwise=CCMP
"""

dnsmasq_conf_content = """interface=lo,ap0
no-dhcp-interface=lo,wlan0
bind-interfaces
server=8.8.8.8
#domain-needed
#bogus-priv
dhcp-range=192.168.10.50,192.168.10.150,12h
address=/majordom-hub/192.168.10.1
"""

wpa_supplicant_content = (
    lambda country: f"""country={country}
ctrl_interface=DIR=/var/run/wpa_supplicant GROUP=netdev
ap_scan=1
update_config=1
"""
)

hosts_content = """
127.0.0.1\tlocalhost
::1\tlocalhost ip6-localhost ip6-loopback
ff02::1\tip6-allnodes
ff02::2\tip6-allrouters
"""

def write(file: str, content: str):
    with open(file, "w") as f:
        f.write(content)

def write_interfaces():
    write(interfaces_path, network_interfaces_content)

def write_default_hostapd():
    write(default_hostapd_path, default_hostapd_content)

def write_hostapd_conf():
    write(hostapd_conf_path, hostapd_conf_content)

def write_dnsmasq_conf():
    write(dnsmaq_conf_path, dnsmasq_conf_content)

def write_wpa_supplicant(country: str = "GB"):
    write(wpa_supplicant_path, wpa_supplicant_content(country))

def write_hosts():
    write(hosts_path, hosts_content)

def write_all():
    write_interfaces()
    write_default_hostapd()
    write_hostapd_conf()
    write_dnsmasq_conf()
    write_wpa_supplicant()
    write_hosts()



from .tools import call_subprocess


def is_hotspot_up() -> bool:
    return call_subprocess('ip link show ap0 | grep "state UP"')

def start_hotspot() -> bool:
    commands = [
        "sudo service hostapd stop",
        "sudo service dnsmasq stop",
        "sudo service dhcpcd stop",
        "sudo iw dev wlan0 interface add ap0 type __ap",
        "sudo service hostapd start",
        "sudo service dnsmasq start",
        "sudo service dhcpcd start",
    ]

    if call_subprocess("; ".join(commands)):
        return is_hotspot_up()
    else:
        return False

def stop_hotspot() -> bool:
    if call_subprocess("sudo service hostapd stop"):
        return not is_hotspot_up()
    else:
        return False



from .tools import call_subprocess


def is_wlan_connected() -> bool:
    return call_subprocess("iwgetid")

def is_eth_connected() -> bool:
    return call_subprocess("ifconfig eth0")

def is_wlan_global_connected() -> bool:
    return call_subprocess("ping -I wlan0 -c 1 google.com")

def is_eth_global_connected() -> bool:
    return call_subprocess("ping -I eth0 -c 1 google.com")

def is_wps_running() -> bool:
    return (
        call_subprocess(
            'wpa_cli -i wlan0 wps_status | grep -o "WPS-STATE=[a-zA-Z]*" | cut -d"=" -f2'
        )
        == "running"
    )



import re
from dataclasses import dataclass
from subprocess import CalledProcessError
from typing import Generator

from .configuration import (
    interfaces_path,
    network_interfaces_content,
    wpa_supplicant_content,
    wpa_supplicant_path,
)
from .tools import call_subprocess, check_subprocess_output


# Variables

country: str = 'GB'

# String parsing

network_regex = re.compile(
    r'network=\{\s+ssid="(?P<ssid>.*)"\s+psk="(?P<psk>.*)"\s+id_str="(?P<id_str>.*)"\s+\}'
)

cell_regex = re.compile(
    r'Cell(?:.|\n)*?Frequency:(?P<frequency>.+)(?:.|\n)*?Quality=(?P<quality>.+)(?:.|\n)*?Encryption key:(?P<encrypted>.+)(?:.|\n)*?ESSID:"(?P<ssid>.+)"'
)

network_record = (
    lambda network, i: f"""
network={{
        ssid="{network.ssid}"
        psk="{network.psk}"
        id_str="AP{i}"
}}
"""
)

# Models

class InterfaceBusyException(Exception):
    pass

@dataclass
class Cell:
    ssid: str
    quality: float
    frequency: str
    encrypted: bool
    # encryption_type: str

@dataclass
class Network:
    ssid: str
    psk: str
    id_str: str = ""

# Functions

def read_networks() -> list[Network]:
    networks: list[Network] = []
    with open(wpa_supplicant_path, "r") as f:
        for match in network_regex.finditer(f.read()):
            networks.append(Network(**match.groupdict()))
    return networks

def write_networks(networks: list[Network]):
    with open(wpa_supplicant_path, "w") as f:
        f.write(wpa_supplicant_content(country))

        for i, network in enumerate(networks):
            f.write(network_record(network, i))

    with open(interfaces_path, "w") as f:
        f.write(network_interfaces_content)

        for i in range(len(networks)):
            f.write(f"iface AP{i} inet dhcp")

def add_network(ssid: str, psk: str):
    networks = read_networks()
    networks.append(Network(ssid, psk))
    write_networks(networks)

def reconnect() -> bool:
    return call_subprocess("sudo wpa_cli -i wlan0 reconfigure")

def scan() -> Generator[Cell, None, None]:
    try:
        output = check_subprocess_output("sudo iwlist wlan0 scan")
    except CalledProcessError as exc:
        raise InterfaceBusyException from exc

    for match in cell_regex.finditer(output):
        groups = match.groupdict()

        quality, max = map(float, groups["quality"].split(" ")[0].split("/"))

        ssid = groups["ssid"].strip()
        quality = round(quality / max, 2)
        frequency = re.sub(r"\(Channel.*\)", "", groups["frequency"]).strip()
        encrypted = groups["encrypted"] == "on"

        yield Cell(ssid, quality, frequency, encrypted)

def start_wps() -> bool:
    return call_subprocess("wpa_cli -i wlan0 wps_pbc")



from . import configuration
from . import hostname
from . import hotspot
from . import status
from . import wifi


def start_hotspot_if_needed():
    connected = status.is_wlan_connected()
    hotspot_up = hotspot.is_hotspot_up()
    
    if not connected and not hotspot_up:
        hotspot.start_hotspot()
    elif connected and hotspot_up:
        hotspot.stop_hotspot()



import time

from .configuration import hosts_content, hosts_path
from .tools import call_subprocess


def set_hostname(hostname: str):
    call_subprocess(f"sudo hostnamectl set-hostname {hostname}")

    with open(hosts_path, "w") as f:
        f.write(hosts_content)
        f.write(f"127.0.1.1 {hostname} {hostname}.localdomain {hostname}.local\n")

    return hostname

def restart_interfaces():  # apply new hostname without reboot
    call_subprocess("sudo ifdown wlan0")
    call_subprocess("sudo ifdown eth0")
    call_subprocess("sudo ifdown ap0")
    time.sleep(1)
    call_subprocess("sudo ifup wlan0")
    call_subprocess("sudo ifup eth0")
    call_subprocess("sudo ifdown ap0")
